#include "stdafx.h"
#include "resource.h"

#define COMPONENT_NAME    "Channel Usage Visualisation"
#define COMPONENT_VERSION "1.0"
#define COMPONENT_ABOUT   "(C) 2005 Holger Stenger"

DECLARE_COMPONENT_VERSION(COMPONENT_NAME, COMPONENT_VERSION, COMPONENT_ABOUT)

CComModule _Module;

struct t_channel_info
{
	t_uint32 m_flag;
	UINT m_id;
	const char * m_name;
};

static const t_channel_info g_channel_info_table[] =
{
	{audio_chunk::channel_back_center,        IDC_CHANNEL_BC,  "Back center"},
	{audio_chunk::channel_back_left,          IDC_CHANNEL_BL,  "Back left"},
	{audio_chunk::channel_back_right,         IDC_CHANNEL_BR,  "Back right"},
	{audio_chunk::channel_front_center,       IDC_CHANNEL_FC,  "Front center"},
	{audio_chunk::channel_front_center_left,  IDC_CHANNEL_FCL, "Front center left"},
	{audio_chunk::channel_front_center_right, IDC_CHANNEL_FCR, "Front center right"},
	{audio_chunk::channel_front_left,         IDC_CHANNEL_FL,  "Front left"},
	{audio_chunk::channel_front_right,        IDC_CHANNEL_FR,  "Front right"},
	{audio_chunk::channel_lfe,                IDC_CHANNEL_LFE, "Low frequency effects"},
	{audio_chunk::channel_side_left,          IDC_CHANNEL_SL,  "Side left"},
	{audio_chunk::channel_side_right,         IDC_CHANNEL_SR,  "Side right"},
	{audio_chunk::channel_top_back_center,    IDC_CHANNEL_TBC, "Top back center"},
	{audio_chunk::channel_top_back_left,      IDC_CHANNEL_TBL, "Top back left"},
	{audio_chunk::channel_top_back_right,     IDC_CHANNEL_TBR, "Top back right"},
	{audio_chunk::channel_top_center,         IDC_CHANNEL_TC,  "Top center"},
	{audio_chunk::channel_top_front_center,   IDC_CHANNEL_TFC, "Top front center"},
	{audio_chunk::channel_top_front_left,     IDC_CHANNEL_TFL, "Top front left"},
	{audio_chunk::channel_top_front_right,    IDC_CHANNEL_TFR, "Top front right"},
};
static const unsigned g_channel_info_table_size = tabsize(g_channel_info_table);

class CChannelUsageDialog :
	public CDialogImpl<CChannelUsageDialog>,
	public message_filter
{
public:
	enum {IDD = IDD_CHANNELVIEW};

	BEGIN_MSG_MAP(CChannelUsageDialog)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_DESTROY(OnDestroy)
		MSG_WM_CLOSE(OnClose)
		COMMAND_HANDLER_EX(IDCANCEL, BN_CLICKED, OnClickedCancel)
		COMMAND_HANDLER_EX(IDC_HIDE_INACTIVE_CHANNELS, BN_CLICKED, OnClickedHideInactiveChannels)
		MSG_WM_TIMER(OnTimer)
		MSG_WM_CTLCOLORSTATIC(OnCtlColorStatic)
	END_MSG_MAP()

	virtual bool pretranslate_message(MSG * p_msg);

private:
	enum {ID_UPDATE_TIMER = 8153};

	virtual void OnFinalMessage(HWND hWnd);
	BOOL OnInitDialog(HWND hWnd, LPARAM lParam);
	void OnDestroy();
	void OnClose();
	void OnClickedCancel(UINT nCode, int nID, HWND hWnd);
	void OnClickedHideInactiveChannels(UINT nCode, int nID, HWND hWnd);
	void OnTimer(UINT nIdEvent, TIMERPROC lpTimerProc);
	HBRUSH OnCtlColorStatic(HDC hDC, HWND hWnd);

	bool get_config_name(pfc::string_base & p_out, unsigned p_count, unsigned p_config);
	void UpdateChannelView(unsigned p_count, unsigned p_config, bool p_force);

	service_ptr_t<visualisation_stream> m_stream;

	CToolTipCtrl m_tooltip;
	unsigned m_last_channel_count, m_last_channel_config;
};

static CWindow g_dlg;

static const GUID guid_cfg_hide_inactive_channels = { 0xb488bd2f, 0x657e, 0x486f, { 0x82, 0x1b, 0x2a, 0x86, 0x99, 0x38, 0xdd, 0x4f } };
static cfg_bool cfg_hide_inactive_channels(guid_cfg_hide_inactive_channels, false);

bool CChannelUsageDialog::pretranslate_message(MSG * p_msg)
{
	/*
	if (p_msg->message == WM_LBUTTONDOWN ||
		p_msg->message == WM_LBUTTONUP ||
		p_msg->message == WM_MOUSEMOVE)
	{
		if (m_tooltip.IsWindow())
			m_tooltip.RelayEvent(p_msg);
	}
*/
	if (IsDialogMessage(p_msg))
		return true;

	return false;
}

void CChannelUsageDialog::OnFinalMessage(HWND hWnd)
{
	delete this;
}

BOOL CChannelUsageDialog::OnInitDialog(HWND hWnd, LPARAM lParam)
{
	m_last_channel_count = m_last_channel_config = 0;

	static_api_ptr_t<message_loop>()->add_message_filter(this);

	UpdateChannelView(0, 0, true);

	CButton btn(GetDlgItem(IDC_HIDE_INACTIVE_CHANNELS));
	btn.SetCheck(cfg_hide_inactive_channels ? BST_CHECKED : BST_UNCHECKED);

	// create visualtion stream
	try
	{
		static_api_ptr_t<visualisation_manager> visman;
		visman->create_stream(m_stream, 0);
	}
	catch (const std::bad_alloc &)
	{
		DestroyWindow();
		return 0;
	}
	catch (const exception_io & exc)
	{
		popup_message::g_show(
			pfc::string8()
				<< "An exception occurred while setting up the visualisation stream:\n"
				<< exc,
			"Channel usage", popup_message::icon_error);
		DestroyWindow();
		return 0;
	}

	// Set timer to 30 FPS
	SetTimer(ID_UPDATE_TIMER, 1000 / 30);

	m_tooltip.Create(m_hWnd);
	for (unsigned n = 0; n < g_channel_info_table_size; n++)
	{
		const t_channel_info & info = g_channel_info_table[n];
		CWindow wnd(GetDlgItem(info.m_id));

		pfc::stringcvt::string_os_from_utf8 osname(info.m_name);
		CRect rc;
		wnd.GetWindowRect(&rc);
		::MapWindowPoints(HWND_DESKTOP, m_hWnd, (LPPOINT)&rc, 2);
		CToolInfo ti(TTF_SUBCLASS, m_hWnd, info.m_id, &rc, const_cast<LPTSTR>(osname.get_ptr()));
		BOOL rv = m_tooltip.AddTool(&ti);
	}

	return TRUE;
}

void CChannelUsageDialog::OnDestroy()
{
	KillTimer(ID_UPDATE_TIMER);

	static_api_ptr_t<message_loop>()->remove_message_filter(this);
}

void CChannelUsageDialog::OnClose()
{
	DestroyWindow();
}

void CChannelUsageDialog::OnClickedCancel(UINT nCode, int nID, HWND hWnd)
{
	DestroyWindow();
}

void CChannelUsageDialog::OnClickedHideInactiveChannels(UINT nCode, int nID, HWND hWnd)
{
	CButton btn(GetDlgItem(IDC_HIDE_INACTIVE_CHANNELS));
	cfg_hide_inactive_channels = btn.GetCheck() != BST_UNCHECKED;
	UpdateChannelView(m_last_channel_count, m_last_channel_config, true);
}

void CChannelUsageDialog::OnTimer(UINT nIdEvent, TIMERPROC lpTimerProc)
{
	try
	{
		if (m_stream.is_empty()) return;

		double t_abs;
		m_stream->get_absolute_time(t_abs);

		audio_chunk_impl chunk;
		m_stream->get_chunk_absolute(chunk, t_abs, 0.01);

		UpdateChannelView(chunk.get_channels(), chunk.get_channel_config(), false);
	}
	catch (const exception_io &)
	{
	}
}

HBRUSH CChannelUsageDialog::OnCtlColorStatic(HDC hDC, HWND hWnd)
{
	UINT nId = ::GetDlgCtrlID(hWnd);
	bool found = false;
	for (unsigned n = 0; n < g_channel_info_table_size; n++)
	{
		if (nId == g_channel_info_table[n].m_id)
		{
			found = true;
			break;
		}
	}

	if (!found || !::IsWindowEnabled(hWnd))
	{
		SetMsgHandled(FALSE);
		return NULL;
	}

	::SetBkColor(hDC, ::GetSysColor(COLOR_HIGHLIGHT));
	::SetTextColor(hDC, ::GetSysColor(COLOR_HIGHLIGHTTEXT));
	return GetSysColorBrush(COLOR_HIGHLIGHT);
}

bool CChannelUsageDialog::get_config_name(pfc::string_base & p_out, unsigned p_count, unsigned p_config)
{
	pfc::string8_fastalloc fmt;
	if (p_count == 0)
		return false;
	if (p_config == 0)
		fmt << "Unknown ";
	else if (p_config == audio_chunk::g_guess_channel_config(p_count))
		fmt << "Default ";
	else
		fmt << "Custom ";
	bool has_lfe = (p_config & audio_chunk::channel_lfe) != 0;
	if (!has_lfe && p_count == 1)
		fmt << "mono";
	if (!has_lfe && p_count == 2)
		fmt << "stereo";
	else if (has_lfe)
		fmt << pfc::format_uint(p_count-1) << ".1";
	else
		fmt << pfc::format_uint(p_count) << ".0";
	p_out = fmt;
	return true;
}

void CChannelUsageDialog::UpdateChannelView(unsigned p_count, unsigned p_config, bool p_force)
{
	if (p_force || m_last_channel_count != p_count || m_last_channel_config != p_config)
	{
		pfc::string8 caption;
		if (!get_config_name(caption, p_count, p_config))
			caption = "Channel usage";
		SetWindowText(pfc::stringcvt::string_os_from_utf8(caption));

		m_last_channel_count = p_count;
		m_last_channel_config = p_config;

		for (unsigned n = 0; n < g_channel_info_table_size; n++)
		{
			const t_channel_info & info = g_channel_info_table[n];
			CWindow wnd(GetDlgItem(info.m_id));

			bool active = (p_config & info.m_flag) != 0;
			wnd.EnableWindow(active);
			wnd.ShowWindow(active ? SW_SHOW : (cfg_hide_inactive_channels ? SW_HIDE : SW_SHOW));
		}
	}
}

static void g_activate()
{
	if (!core_api::assert_main_thread()) return;

	if (!g_dlg.IsWindow())
	{
		CChannelUsageDialog * dlg = new CChannelUsageDialog();
		g_dlg = dlg->Create(core_api::get_main_window());
	}
	if (g_dlg.IsWindow())
	{
		g_dlg.ShowWindow(SW_SHOW);
		::SetForegroundWindow(g_dlg);
	}
}

static const GUID guid_main_channelusage = { 0xff8c443d, 0x4eea, 0x4ffd, { 0xa3, 0x68, 0x43, 0xde, 0x38, 0xce, 0x8b, 0xe3 } };

class mainmenu_commands_channelusage : public mainmenu_commands
{
public:
	virtual t_uint32 get_command_count()
	{
		return 1;
	}

	virtual GUID get_command(t_uint32 p_index)
	{
		if (p_index == 0)
			return guid_main_channelusage;
		return pfc::guid_null;
	}

	virtual void get_name(t_uint32 p_index,pfc::string_base & p_out)
	{
		if (p_index == 0)
			p_out = "Channel Usage";
	}

	virtual bool get_description(t_uint32 p_index,pfc::string_base & p_out)
	{
		return false;
	}

	virtual GUID get_parent()
	{
		return mainmenu_groups::view_visualisations;
	}

	virtual t_uint32 get_sort_priority()
	{
		return sort_priority_dontcare;
	}

	virtual void execute(t_uint32 p_index,service_ptr_t<service_base> p_callback)
	{
		if (p_index == 0)
			g_activate();
	}
};

static mainmenu_commands_factory_t<mainmenu_commands_channelusage> foo_mainmenu_commands;
